
/********************************************************
* COS265 Data Structures & Algorithms                   *
* Project 5 - Small Group kD Trees                      *
* by Chad Jordan 11/16/07                               *
*                                                       *
* Java implementation for a point and search            *
* bruteforce algorithm on a 2D plane.                   *
* Dependencies: none                                    *
********************************************************/
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

public class BruteForceSearch<Point> {

    private List<Point> points;

    public BruteForceSearch() {
        points = new ArrayList<>();
    }

    public void insert(Point point) {
        points.add(point);
    }

    public Point nearest(Point query) {
        if (points.isEmpty()) {
            return null;
        }

        Point nearestPoint = points.get(0);
        double minDistance = distance(nearestPoint, query);

        for (Point point : points) {
            double distance = distance(point, query);
            if (distance < minDistance) {
                minDistance = distance;
                nearestPoint = point;
            }
        }

        return nearestPoint;
    }
    
    class NestedClass RedBlackBST<Point> {

        private static final boolean RED = true;
        private static final boolean BLACK = false;
    
        private Node root;
    
        private class Node {
            private Point point;
            private Node left, right;
            private int size;
            private boolean color;
    
            public Node(Point point, boolean color, int size) {
                this.point = point;
                this.color = color;
                this.size = size;
            }
        }
    
        public RedBlackBST() {
        }
    
        // is node red; false if node is null
        private boolean isRed(Node x) {
            if (x == null) return false;
            return x.color == RED;
        }
    
        // number of nodes in subtree rooted at x; 0 if x is null
        private int size(Node x) {
            if (x == null) return 0;
            return x.size;
        }
    
        // return number of points
        public int size() {
            return size(root);
        }
    
        public boolean isEmpty() {
            return root == null;
        }
    
        // insert the point
        public void insert(Point point) {
            root = insert(root, point);
            root.color = BLACK;
        }
    
        // insert the point into the tree rooted at h
        private Node insert(Node h, Point point) {
            if (h == null) return new Node(point, RED, 1);
    
            int cmp = comparePoints(point, h.point);
            if (cmp < 0) {
                h.left = insert(h.left, point);
            } else if (cmp > 0) {
                h.right = insert(h.right, point);
            }
    
            // fix-up any right-leaning links
            if (isRed(h.right) && !isRed(h.left)) h = rotateLeft(h);
            if (isRed(h.left) && isRed(h.left.left)) h = rotateRight(h);
            if (isRed(h.left) && isRed(h.right)) flipColors(h);
    
            h.size = size(h.left) + size(h.right) + 1;
    
            return h;
        }
    
        // compare two points
        private int comparePoints(Point p1, Point p2) {
            double dx = ((Point2D.Double) p1).getX() - ((Point2D.Double) p2).getX();
            double dy = ((Point2D.Double) p1).getY() - ((Point2D.Double) p2).getY();
            if (dx < 0) return -1;
            else if (dx > 0) return 1;
            else if (dy < 0) return -1;
            else if (dy > 0) return 1;
            else return 0;
        }
    
        // rotate right
        private Node rotateRight(Node h) {
            Node x = h.left;
            h.left = x.right;
            x.right = h;
            x.color = x.right.color;
            x.right.color = RED;
            x.size = h.size;
            h.size = size(h.left) + size(h.right) + 1;
            return x;
        }
    
        // rotate left
        private Node rotateLeft(Node h) {
            Node x = h.right;
            h.right = x.left;
            x.left = h;
            x.color = x.left.color;
            x.left.color = RED;
            x.size = h.size;
            h.size = size(h.left) + size(h.right) + 1;
            return x;
        }
    
        // flip colors
        private void flipColors(Node h) {
            h.color = !h.color;
            h.left.color = !h.left.color;
            h.right.color = !h.right.color;
        }
    
        // nearest point
        public Point nearest(Point query) {
            if (root == null) return null;
            return nearest(root, query, root.point);
        }
    
        private Point nearest(Node x, Point query, Point nearest) {
            if (x == null) return nearest;
    
            if (comparePoints(query, x.point) < 0) {
                nearest = nearest(x.left, query, nearest);
                if (x.right != null && comparePoints(query, x.right.point) < 0) {
                    nearest = nearest(x.right, query, nearest);
                }
            } else {
                nearest = nearest(x.right, query, nearest);
                if (x.left != null && comparePoints(query, x.left.point) >= 0) {
                    nearest = nearest(x.left, query, nearest);
                }
            }
    
            return nearest;
        }
        
        // Compare two points lexicographically
    private int comparePoints(Point2D.Double p1, Point2D.Double p2) {
        if (p1.getX() < p2.getX()) return -1;
        else if (p1.getX() > p2.getX()) return 1;
        else if (p1.getY() < p2.getY()) return -1;
        else if (p1.getY() > p2.getY()) return 1;
        else return 0;
    }

    // Calculate Euclidean distance between two points
    private double distance(Point2D.Double p1, Point2D.Double p2) {
        return Math.sqrt(Math.pow(p2.getX() - p1.getX(), 2) + Math.pow(p2.getY() - p1.getY(), 2));
    }
    private double distance(Point p1, Point p2) {
        // Assuming points are of type Point2D.Double
        double dx = ((java.awt.geom.Point2D.Double) p2).getX() - ((java.awt.geom.Point2D.Double) p1).getX();
        double dy = ((java.awt.geom.Point2D.Double) p2).getY() - ((java.awt.geom.Point2D.Double) p1).getY();
        return Math.sqrt(dx * dx + dy * dy);
    }

    public static void main(String[] args) {
        BruteForceSearch<java.awt.geom.Point2D.Double> bruteForceSearch = new BruteForceSearch<>();
        RedBlackBST<Point2D.Double> redBlackBST = new RedBlackBST<>();
    

        // Insert some points
        bruteForceSearch.insert(new java.awt.geom.Point2D.Double(2, 3));
        bruteForceSearch.insert(new java.awt.geom.Point2D.Double(5, 4));
        bruteForceSearch.insert(new java.awt.geom.Point2D.Double(9, 6));
        bruteForceSearch.insert(new java.awt.geom.Point2D.Double(4, 7));

        // Query for nearest point
        java.awt.geom.Point2D.Double query = new java.awt.geom.Point2D.Double(6, 5);
        java.awt.geom.Point2D.Double nearest = bruteForceSearch.nearest(query);

        System.out.println("Nearest point to " + query + ": " + nearest);
    }
}
}