/************************************************************
* COS 265 Data Structures & Algorithms                      *
* Project 5 - Small Group kD Trees                          *
* by Chad Jordan  11/16/07                                  *
*                                                           *
* Java implementation of an algorithm to find the nearest   *
* neighbors in a 2D plane.                                  *
* Dependencies: none                                        *
************************************************************/
import java.util.Comparator;

public class TwoDTree<Key extends Comparable<Key>, Value> {
    private Node root;

    private class Node {
        private Key key;
        private Value value;
        private Node left, right;
        private int size; // Number of nodes in subtree

        public Node(Key key, Value value, int size) {
            this.key = key;
            this.value = value;
            this.size = size;
        }
    }

    // Compare points by their x-coordinate
    private Comparator<Key> xComparator = Comparator.comparingDouble(o -> ((Point) o).getX());

    // Compare points by their y-coordinate
    private Comparator<Key> yComparator = Comparator.comparingDouble(o -> ((Point) o).getY());

    // Symbol table methods
    public boolean isEmpty() {
        return size() == 0;
    }

    public int size() {
        return size(root);
    }

    private int size(Node x) {
        if (x == null) return 0;
        return x.size;
    }

    public void put(Key key, Value value) {
        if (key == null) throw new IllegalArgumentException("Key cannot be null");
        root = put(root, key, value, xComparator, 0);
    }

    private Node put(Node x, Key key, Value value, Comparator<Key> comparator, int depth) {
        if (x == null) return new Node(key, value, 1);

        int cmp = comparator.compare(key, x.key);
        if (cmp < 0) {
            x.left = put(x.left, key, value, depth % 2 == 0 ? yComparator : xComparator, depth + 1);
        } else if (cmp > 0) {
            x.right = put(x.right, key, value, depth % 2 == 0 ? yComparator : xComparator, depth + 1);
        } else {
            x.value = value;
        }

        x.size = 1 + size(x.left) + size(x.right);
        return x;
    }

    public Value get(Key key) {
        return get(root, key, xComparator);
    }

    private Value get(Node x, Key key, Comparator<Key> comparator) {
        while (x != null) {
            int cmp = comparator.compare(key, x.key);
            if (cmp < 0) {
                x = x.left;
            } else if (cmp > 0) {
                x = x.right;
            } else {
                return x.value;
            }
        }
        return null;
    }

    // Nearest neighbor search
    public Value nearestNeighbor(Point query) {
        return nearestNeighbor(root, query, root.key, xComparator);
    }

    private Value nearestNeighbor(Node x, Point query, Key nearest, Comparator<Key> comparator) {
        if (x == null) return null;

        double currentDistance = distance(query, (Point) x.key);
        double bestDistance = distance(query, (Point) nearest);

        if (currentDistance < bestDistance) {
            nearest = x.key;
            bestDistance = currentDistance;
        }

        int cmp = comparator.compare(query, x.key);
        if (cmp < 0) {
            Value leftValue = nearestNeighbor(x.left, query, nearest, comparator == xComparator ? yComparator : xComparator);
            if (leftValue != null && distance(query, (Point) leftValue) < bestDistance) {
                nearest = (Key) leftValue;
                bestDistance = distance(query, (Point) leftValue);
            }
            if (x.right != null && Math.abs(comparator.compare(query, x.right.key)) < bestDistance) {
                Value rightValue = nearestNeighbor(x.right, query, nearest, comparator == xComparator ? yComparator : xComparator);
                if (rightValue != null && distance(query, (Point) rightValue) < bestDistance) {
                    nearest = (Key) rightValue;
                    bestDistance = distance(query, (Point) rightValue);
                }
            }
        } else {
            Value rightValue = nearestNeighbor(x.right, query, nearest, comparator == xComparator ? yComparator : xComparator);
            if (rightValue != null && distance(query, (Point) rightValue) < bestDistance) {
                nearest = (Key) rightValue;
                bestDistance = distance(query, (Point) rightValue);
            }
            if (x.left != null && Math.abs(comparator.compare(query, x.left.key)) < bestDistance) {
                Value leftValue = nearestNeighbor(x.left, query, nearest, comparator == xComparator ? yComparator : xComparator);
                if (leftValue != null && distance(query, (Point) leftValue) < bestDistance) {
                    nearest = (Key) leftValue;
                    bestDistance = distance(query, (Point) leftValue);
                }
            }
        }

        return (Value) nearest;
    }

    private double distance(Point p1, Point p2) {
        return Math.sqrt(Math.pow(p2.getX() - p1.getX(), 2) + Math.pow(p2.getY() - p1.getY(), 2));
    }

    // Test the TwoDTree
    public static void main(String[] args) {
        TwoDTree<Point, String> tree = new TwoDTree<>();
        tree.put(new Point(2, 3), "A");
        tree.put(new Point(5, 4), "B");
        tree.put(new Point(9, 6), "C");
        tree.put(new Point(4, 7), "D");

        Point query = new Point(6, 5);
        System.out.println("Nearest neighbor to " + query + ": " + tree.nearestNeighbor(query));
    }
}

class Point {
    private double x;
    private double y;

    public Point(double x, double y) {
        this.x = x;
        this.y = y;
    }

    public double getX() {
        return x;
    }

    public double getY() {
        return y;
    }

    @Override
    public String toString() {
        return "(" + x + ", " + y + ")";
    }
}